/*  Copyright 2011-2013 Alexis Herault, Giuseppe Bilotta, Robert A. Dalrymple, Eugenio Rustico, Ciro Del Negro

    Istituto Nazionale di Geofisica e Vulcanologia
        Sezione di Catania, Catania, Italy

    Università di Catania, Catania, Italy

    Johns Hopkins University, Baltimore, MD

    This file is part of GPUSPH.

    GPUSPH is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    GPUSPH is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GPUSPH.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <iostream>
#include <stdexcept>

#include "GlobalData.h"
#include "cudasimframework.cu"

#include "WindTunnel.h"


#define INLET_VELOCITY 14.


WindTunnel::WindTunnel(GlobalData *_gdata) : XProblem(_gdata)
{
    // Size and origin of the simulation domain
    lx = 4.;
    ly = 3.;
    lz = 3.;
    m_size = make_double3(lx, ly, lz);
    m_origin = make_double3(0, 0, 0);

    SETUP_FRAMEWORK(
		kernel<WENDLAND>,
		formulation<SPH_F1>,
		viscosity<DYNAMICVISC>,
        boundary<SA_BOUNDARY>,
		periodicity<PERIODIC_NONE>,
        flags<ENABLE_INLET_OUTLET | ENABLE_DENSITY_SUM | ENABLE_DTADAPT | ENABLE_FERRARI>
    );

    addFilter(SHEPARD_FILTER, 20); // or MLS_FILTER

    // SPH parameters
    set_deltap(0.1f);;
    simparams()->dt = 0.001; //NOTE originally was 0.0001
    simparams()->dtadaptfactor = 0.2;
    simparams()->buildneibsfreq = 10;
    simparams()->tend = 10.0f; //seconds
    simparams()->maxneibsnum = 640;

	setWaterLevel(2.);
	setMaxParticleSpeed(100.);

    simparams()->ferrari = 1.0f; //seconds
	simparams()->ferrariLengthScale = 0.25f;

    //WaveGage
    if (get_option("gages", false)) {
        add_gage(1, 0.3);
        add_gage(0.5, 0.3);
    }

    // Physical parameters
    H = 0.45;
    physparams()->gravity = make_float3(0.0f, 0.0f, 0.0f);

    float r0 = m_deltap; // m_deltap initial particle spacing defined in Problem.h
    physparams()->r0 = r0;

    // setting air density 1.1845 kg/m^3 at 25deg
    float rho = 1.1845;
    add_fluid(rho); // index will be 0
    // setting air kinematic viscosity 1.225 kg/m^3
    set_kinematic_visc(/*fluid index*/0, 1.5571e-5);
    // setting air equation of state (assuming incompressible behaviour here)
    set_equation_of_state(/*fluid index*/0,  7.0f, 20.f);
     
    soundspeed(rho, /*fluid index*/0);

    pressure(rho, /*fluid index*/0);

    physparams()->artvisccoeff =  0.2;
    physparams()->smagfactor = 0.12*0.12*m_deltap*m_deltap;
    physparams()->kspsfactor = (2.0/3.0)*0.0066*m_deltap*m_deltap;
    physparams()->epsartvisc = 0.01*simparams()->slength*simparams()->slength;

    // BC when using LJ
    physparams()->dcoeff = 0.5*9.81*H;

    addHDF5File(GT_FLUID, Point(0, 0, 0), "../0.fluid.h5sph", NULL);

    addHDF5File(GT_FIXED_BOUNDARY, Point(0, 0, 0), "../0.walls.h5sph", NULL);

    GeometryID inlet = 
        addHDF5File(GT_OPENBOUNDARY, Point(0, 0, 0), "../0.flow_in.h5sph", NULL);
    disableCollisions(inlet);
    setVelocityDriven(inlet, VELOCITY_DRIVEN);

    GeometryID outCtePressure = 
        addHDF5File(GT_OPENBOUNDARY, Point(0, 0, 0), "../0.constant_pressure.h5sph", NULL);
    disableCollisions(outCtePressure);
    setVelocityDriven(outCtePressure, PRESSURE_DRIVEN);

    GeometryID outCteVelocity = 
        addHDF5File(GT_OPENBOUNDARY, Point(0, 0, 0), "../0.flow_out.h5sph", NULL);
    disableCollisions(outCteVelocity);
    setVelocityDriven(outCteVelocity, VELOCITY_DRIVEN);

    //Wave paddle definition:  location, start & stop times, stroke and frequency (2 \pi/period)

    paddle_length = 0.7;
    paddle_width = 0.3;
    paddle_tstart = 0.5f;
    paddle_origin = make_double3(1, 1., 0.3);
    paddle_tend = 30.0f;//seconds
    // The stroke value is given at free surface level H
    float stroke = 0.2;
    // m_mbamplitude is the maximal angular value for paddle angle
    // Paddle angle is in [-m_mbamplitude, m_mbamplitude]
    paddle_amplitude = atan(stroke/(2.0*(H - paddle_origin.z)));
    cout << "\npaddle_amplitude (radians): " << paddle_amplitude << "\n";
    paddle_omega = 2.0*M_PI/0.8;        // period T = 0.8 s

    const float amplitude = -paddle_amplitude ;
    GeometryID paddle = addBox(GT_MOVING_BODY, FT_BORDER, Point(paddle_origin),    0, paddle_width, paddle_length);
    rotate(paddle, 0, -amplitude, 0);
    disableCollisions(paddle);

    // Drawing and saving times
    add_writer(VTKWRITER, .1);  //second argument is saving time in seconds

    // Name of problem used for directory creation
    m_name = "WindTunnel";
}


void WindTunnel::moving_bodies_callback(const uint index, Object* object, const double t0, const double t1,
        const float3& force, const float3& torque, const KinematicData& initial_kdata,
        KinematicData& kdata, double3& dx, EulerParameters& dr)
{
    dx = make_double3(0.0);
    kdata.lvel = make_double3(0.0f, 0.0f, 0.0f);
    if (t1> paddle_tstart && t1 < paddle_tend)
    {
        kdata.avel = make_double3(0.0, paddle_amplitude*paddle_omega*sin(paddle_omega*(t1-paddle_tstart)),0.0);
        EulerParameters dqdt = 0.5*EulerParameters(kdata.avel)*kdata.orientation;
        dr = EulerParameters::Identity() + (t1-t0)*dqdt*kdata.orientation.Inverse();
        dr.Normalize();
        kdata.orientation = kdata.orientation + (t1 - t0)*dqdt;
        kdata.orientation.Normalize();
    }
    else
    {
        kdata.avel = make_double3(0.0,0.0,0.0);
        kdata.orientation = kdata.orientation;
        dr.Identity();
    }
}


void WindTunnel::imposeBoundaryConditionHost(
    MultiBufferList::iterator bufwrite, MultiBufferList::const_iterator bufread,
    uint* IOwaterdepth, const float t,
    const uint numParticles,
    const uint numOpenBoundaries,
    const uint particleRangeEnd)
{
	//float4 *newVel = bufwrite->getData<BUFFER_VEL>();
	float4 *newEulerVel = bufwrite->getData<BUFFER_EULERVEL>();
	//float *newTke = bufwrite->getData<BUFFER_TKE>();
	//float *newEpsilon = bufwrite->getData<BUFFER_EPSILON>();

	//const particleinfo *info = bufread->getData<BUFFER_INFO>();
	//const float4 *oldPos = bufread->getData<BUFFER_POS>();
	//const hashKey *particleHash = bufread->getData<BUFFER_HASH>();

    for (int i=0; i<numParticles; i++) {
        newEulerVel[i].x = INLET_VELOCITY * t;
        //newEulerVel[i].w = RHO(pressure, fluid_num(info);
    }

}

